import os


MODEL_TOKEN_CONFIGS = dict({
    "gpt-3.5-turbo": dict(
        safe_buffer=16,
        hard_roof=4096,
        min_answer_room=512,
    ),
    # gpt4 config here
})

OPENAI_API_KEY = os.environ.get(
    "OPENAI_API_KEY", "not set")



import os
from pathlib import Path
import math
from typing import List, Iterator

import openai
from pydub import AudioSegment

from important.shell.color import cprint

MAX_BYTES = 26000000  # 25MB limit for OpenAI API
MODEL_NAME = "whisper-1"


class Transcribe:
    """
    Transcribe a video file to string
    """

    def __init__(self, video_filepath: str, lang: str, verbose: bool = False):
        self.verbose = verbose
        self.lang = lang
        self.video_filepath = video_filepath
        self.mp4_to_mp3()
        self.audio_size = os.path.getsize(self.audio_filepath)
        self.num_chunks = math.ceil(self.audio_size / MAX_BYTES)

        if self.verbose > 0:
            cprint(f"Language: {self.lang}", color="blue")
            cprint(f"Max size:\t{MAX_BYTES}", color="blue")
            cprint(f"Audio size:\t{self.audio_size}", color="blue")
            cprint(f"Number of chunks:\t{self.num_chunks}", color="blue")

    def get_audio(self):
        return AudioSegment.from_file(self.video_filepath, format=Path(self.video_filepath).suffix[1:])

    @property
    def audio_filepath(self) -> str:
        """
        Returns the filepath of the audio file
        """
        return self.video_filepath.replace(Path(self.video_filepath).suffix, ".mp3")

    def mp4_to_mp3(self) -> None:
        """
        Extracts the audio from the video file and saves it as an mp3
        """
        self.get_audio().export(self.audio_filepath)

    def get_audio_chunk_paths(self) -> Iterator:
        if self.audio_size >= MAX_BYTES:
            audio = self.get_audio()
            chunk_duration_ms = math.ceil(audio.duration_seconds * 1000 / self.num_chunks)
            for i in range(self.num_chunks):
                start_ms = i * chunk_duration_ms
                end_ms = start_ms + chunk_duration_ms
                chunk = audio[start_ms:end_ms]
                chunk_filepath = "__".join([self.audio_filepath.split(".")[0], str(start_ms), str(end_ms)]) + ".mp3"
                chunk.export(chunk_filepath)
                yield chunk_filepath
        else:
            yield self.audio_filepath

    @staticmethod
    def audio_transcribe(chunk_filepath: str, language: str) -> str:
        """
        Transcribes an audio file according to the language
        """
        with open(chunk_filepath, "rb") as audio_file:
            transcript = openai.Audio.transcribe(file=audio_file, model=MODEL_NAME, language=language)
        return transcript.text

    def __call__(self) -> str:
        transcription = ""
        for i, chunk_filepath in enumerate(self.get_audio_chunk_paths()):
            text_chunk = self.audio_transcribe(chunk_filepath, self.lang)

            if self.verbose > 1:
                cprint(f"[{self.lang.upper()}]\tTranscription Chunk {i+1}", color="meganta")
                cprint(text_chunk, color="cyan")

            transcription += text_chunk

        return transcription



from kubernetes_asyncio import client, config
from kubernetes_asyncio.client.api_client import ApiClient
from forgebox.unpack import Unpack
import pandas as pd
from .utils import calc_delta
from typing import Dict, List


async def list_services_async(
    namespace: str
) -> pd.DataFrame:
    """
    Return the service list within the namespace
    The async version
    """
    await config.load_kube_config()

    async with ApiClient() as api_client:
        v1 = client.CoreV1Api(api_client)
        endpoint_res = await v1.list_namespaced_endpoints(namespace=namespace)
        services = list()
        for ep in endpoint_res.items:
            name, ports, version, create_ts = Unpack(ep.to_dict())(
                ["metadata", "name"],
                ["subsets", 0, "ports"],
                ["metadata", "resource_version"],
                ["metadata", "creation_timestamp"]
            )
            services.append(
                dict(
                    name=name,
                    ports=ports,
                    version=version,
                    create_ts=create_ts
                )
            )
        services_df = pd.DataFrame(services)
        services_df['age'] = services_df.create_ts.apply(calc_delta)
        return services_df


async def list_pods_async(
    namespace: str
) -> pd.DataFrame:
    await config.load_kube_config()

    async with ApiClient() as api_client:
        v1 = client.CoreV1Api(api_client)
        k8_res = await v1.list_namespaced_pod(namespace=namespace)
        pods = list()
        for item in k8_res.items:
            data = dict(zip(
                ["name", "version", "app", "created_ts",
                 "image", "ready", "started"],
                Unpack(item.to_dict())(
                    ["metadata", "name"],
                    ["metadata", "resource_version"],
                    ["metadata", "labels", "app"],
                    ["metadata", "creation_timestamp"],
                    ["spec", "containers", 0, "image"],
                    ["status", "container_statuses", 0, "ready"],
                    ["status", "container_statuses", 0, "started"],
                )))
            pods.append(data)
        pods_df = pd.DataFrame(pods)
        pods_df["image_tag"] = (
            pods_df["image"]
            .apply(lambda x: x.split("/")[1])
        )
        pods_df["age"] = pods_df["created_ts"].apply(calc_delta)
        return pods_df


async def services_to_imagetags_async(namespace: str) -> Dict[str, List[str]]:
    """
    Given a service name, return the image tag
    """
    await config.load_kube_config()

    async with ApiClient() as api_client:
        v1 = client.CoreV1Api(api_client)
        k8_res = await v1.list_namespaced_pod(namespace=namespace)
        services = dict()
        for item in k8_res.items:
            name, image = Unpack(item.to_dict())(
                ["metadata", "labels", "app"],
                ["spec", "containers", 0, "image"],
            )
            if name not in services:
                services[name] = list()
            services[name].append(image.split("/")[1])
        return services



import logging
from typing import (
    Dict, Any, Optional
    )
import json


from gallop.config import BaseConfig


EXAMPLE_CONF = """
classes:
  User:
    upsert_by:
      - user_name
  Order:
    upsert_by:
      # we'll update or create the order
      # by the following names of column
      - user_id
      - product_name
objects:
  order1:
    class: Order
    data:
      product_name: pepper spray
      user_id: $user1.id
  order2:
    class: Order
    data:
      product_name: paper clip
      user_id: $user2.id
  user1:
    class: User
    data:
      user_name: jason
      user_email: bourn@gmail.com
  user2:
    class: User
    data:
      user_name: ethan
      user_email: hunt@gmail.com
"""


def base_config_to_dict(config: BaseConfig):
    """
    Convert the config to dict
    """
    return config.to_dict()


FIELD_PROCESS = {
    "json.loads": json.loads,
    "json.dumps": json.dumps,
    "base_config_to_dict": base_config_to_dict,
}


class StatefulBase:
    class_room = dict()

    def __init__(
        self,
        config: BaseConfig,
        class_room: Dict[str, Any] = dict(),
        init_objects: Optional[Dict[str, Any]] = None,
    ):
        self.class_room.update(class_room)
        self.class_confs = config.classes
        self.objects_conf = config.objects
        self.class_confs = self.validate_class(self.class_confs)
        self.objects = dict() if init_objects is None else init_objects
        self.init_order = list()

    def validate_class(self, class_confs: BaseConfig):
        """
        Validate the classes in the config
        """
        for class_name, _ in class_confs.items():
            if class_name not in self.class_room:
                raise KeyError(
                    f"Class {class_name} is not defined in class_room, "
                    f"what we have is {self.class_room.keys()}"
                )
            class_confs[class_name]['class'] = self.class_room[class_name]
        return class_confs

    def __repr__(self) -> str:
        return str(self.objects_conf)

    def init_key_value(self, key: str, value: Any):
        """
        Init the key value
        """
        if type(value) == str and value.startswith('$'):
            variable_chain = value[1:].split('.')
            obj_name = variable_chain[0]
            # dead loop check
            if obj_name == key:
                raise ValueError(
                    f"❌ Cannot reference self '{key}'")
            obj = self[obj_name]
            for attr in variable_chain[1:]:
                obj = getattr(obj, attr)
            value = obj
        return key, value

    def init_value_dict(self, values: Dict[str, Any], ):
        """
        Read the value recursively
        """
        for key, value in values.items():
            key, obj = self.init_key_value(key, value)
            values[key] = obj
        return values

    def __getitem__(self, obj_name: str) -> Any:
        """
        Get the variable by name
        """
        if obj_name in self.objects:
            return self.objects[obj_name]
        if obj_name not in self.objects_conf:
            raise KeyError(
                f"❌ Object {obj_name} is not defined in the objects config, "
                f"what we have is {self.objects_conf.keys()}"
            )
        obj_conf = self.objects_conf[obj_name]
        class_name = obj_conf['class']
        class_conf = self.class_confs[class_name]

        # if there is upsert condition
        # we are going to query with the upsert fields
        if "upsert_by" in class_conf:
            upsert_by = class_conf.upsert_by

            # set up query condition
            query = dict()
            for attr in upsert_by:
                obj_val = obj_conf.data[attr]
                # incase the value is a reference
                obj_val = self.init_key_value(attr, obj_val)[1]
                query[attr] = obj_val
            obj = self.special_query(class_name, **query)
            if obj is not None:
                # update the object properties
                has_read_process = "read_process" in class_conf
                for attr in obj_conf.data:
                    if attr not in upsert_by:
                        obj_val = self.init_key_value(
                            attr, obj_conf.data[attr])[1]
                        if has_read_process:
                            if attr in class_conf.read_process:
                                process = FIELD_PROCESS[
                                    class_conf.read_process[attr]]
                                obj_val = process(obj_val)
                        setattr(obj, attr, obj_val)
                self.special_set(obj_name, obj)

                self.objects[obj_name] = obj
                return obj

        # no exist object detected
        # create new object
        cls = class_conf['class']
        obj = cls(**self.init_value_dict(
            obj_conf['data']))

        # if there is a `read_process` for the object
        if "read_process" in class_conf:
            for field_name in obj_conf.data:
                if field_name in class_conf.read_process:
                    process = FIELD_PROCESS[
                        class_conf.read_process[field_name]]
                    logging.debug(
                        f"Processing {obj_name}.{field_name} "
                        f"with {process}")
                    setattr(
                        obj,
                        field_name,
                        process(getattr(obj, field_name)))

        # we keep track of init order for save the data
        # hierarchically later
        self.init_order.append(obj_name)
        self.objects[obj_name] = obj
        self.special_set(obj_name, obj)
        return obj

    def __setitem__(self, obj_name: str, value: Any) -> None:
        self.objects[obj_name] = value

    def __len__(self) -> int:
        return len(self.objects)

    def __call__(self,) -> Dict[str, Any]:
        for obj_name in self.objects_conf.keys():
            self[obj_name] = self[obj_name]
        return self.objects

    def special_set(self, obj_name: str, obj: Any):
        """
        Set the attribute of the object
            Please override this method to implement
        """
        logging.warning(
            f"Not implemented special_set for {obj_name}")

    def special_query(self, class_name: str, **query) -> Optional[Any]:
        """
        Query the attribute of the object
            Please override this method to implement
        """
        logging.warning(
            f"Not implemented special_query for {class_name} {query}")
        return None


def build_db_stateful(
    db: "Session",
    base: "Base",
) -> StatefulBase:
    """
    Build a stateful class for database connection
    """
    class_room_db = dict(
        (i.class_.__name__, i.class_)
        for i in base.registry.mappers)

    class DBStateful(StatefulBase):
        """
        Stateful class for database
        """
        class_room = class_room_db

        def special_set(
            self,
            obj_name: str,
            obj: Any
        ):
            """
            set object to database
            """
            db.add(obj)
            db.commit()
            db.refresh(obj)

        def special_query(self, class_name: str, **query) -> Optional[Any]:
            """
            Query the object from database
            """
            cls = self.class_confs[class_name]['class']
            return db.query(cls).filter_by(**query).first()

        def delete(self, obj_name: str):
            """
            Delete the object from database
            """
            obj = self[obj_name]
            db.delete(obj)
            db.commit()
    return DBStateful



import pandas as pd
from subprocess import run
from io import StringIO
from typing import Union
import logging
from traceback import format_exc
from .async_pandas import read_cmd_async


def read_cmd(cmd: str, sep='\t') -> Union[pd.DataFrame, str]:
    """
    Parse the command result to a pandas dataframe
    """
    run_res = run(
        cmd,
        shell=True,
        capture_output=True,
        )
    if run_res.returncode != 0:
        error_msg = run_res.stderr.decode('utf-8')
        logging.error(error_msg)
        return error_msg
    value = run_res.stdout.decode('utf-8')
    output = StringIO(
        value
    )
    try:
        df = pd.read_csv(output, sep=sep)
        return df
    except KeyboardInterrupt as e:
        raise e
    except Exception:
        logging.error(format_exc())
        return value



import os
import logging
from subprocess import call
from important.redis import (
    RedisController, RedisControllerAsync,
    CallableDict)
from important.location import CONF_DIR
from redis import Redis
from redis.asyncio import Redis as RedisAsync
import asyncio
from time import sleep


logging.info(
    "Make sure you have ran the 'test/start-env.sh' " +
    "before you start this test")


os.environ['REDIS_HOST'] = "localhost"


redis_client = Redis(host="localhost", db=0)
redis_client_async = RedisAsync(host="localhost", db=0)


control = RedisController.from_task(redis_client, task="channel_pop")


logging.debug(control)


def test_ttl_to_seconds():
    # let's try to reconstruct the ttl description to seconds
    assert RedisController.ttl_to_seconds(
        months=1, days=1, hours=1, minutes=1, seconds=1) == 2682061
    assert RedisController.ttl_to_seconds(
        minutes=1, seconds=1) == 61
    assert RedisController.ttl_to_seconds(
        hours=1, seconds=1) == 3601


def test_mk_key():
    channel_id = 1776
    user_id = 123123

    key = control.mk_key(channel_id=channel_id, user_id=user_id)
    assert key == f"chnl_pop:{channel_id}"
    try:
        key = control.mk_key(user_id=user_id)
        raise NotImplementedError("Failed to raise KeyError")
    except KeyError:
        logging.info("No worry, the above keyerror is a calculated error")
        assert True
    except Exception:
        raise NotImplementedError("Failed to raise KeyError")


def test_set_get():
    control[dict(channel_id=123123)] = 'dummy result'
    assert control[dict(channel_id=123123)] == 'dummy result'


def test_async_set_get():
    async def async_test():
        contorl = RedisControllerAsync.from_task(
            redis_client_async, task="channel_pop")
        await contorl.set(dict(channel_id=123123), 'dummy result')
        assert await contorl[dict(channel_id=123123)] == 'dummy result'
        # test the ttl
        logging.info("Let's count out 2.5 seconds to expire the key")
        sleep(2.5)
        assert await contorl[dict(channel_id=123123)] is None
    asyncio.run(async_test())


redis_client.set("some_global", 1, ex=10)


def func_builder(
    in_func_arg1,
    in_func_arg2,
    in_func_key1=None,
    in_func_key2=1
):
    logging.debug(f"loading func_builder with {locals()}")

    def func(*args, **kwargs):
        return dict(
            in_func_arg1=in_func_arg1,
            in_func_arg2=in_func_arg2,
            in_func_key1=in_func_key1,
            in_func_key2=in_func_key2,
            args=args,
            kwargs=kwargs,
            some_global=int(redis_client.get("some_global"))
        )
    return func


def test_callable_dict(caplog):
    callable_dict = CallableDict(
        client=redis_client,
        class_map=dict(func_builder=func_builder),
        delay=1,
    )
    for i in range(20):
        callable_dict.set("func_builder",
                          "key1", "key2",
                          in_func_key1="key3",
                          in_func_key2="key3")
        callable = callable_dict.get("func_builder",
                                     "key1", "key2",
                                     in_func_key1="key3",
                                     in_func_key2="key3")

        result = callable(i+1)

        assert result == dict(
            in_func_arg1="key1",
            in_func_arg2="key2",
            in_func_key1="key3",
            in_func_key2="key3",
            args=(i+1,),
            kwargs={},
            some_global=1,
        )

    # no redis checking
    assert str(caplog.text).count("🍄") == 0
    # 1 initialization
    assert str(caplog.text).count("✨") == 1

    sleep(1.1)
    for i in range(2):  # to test the delay prolong is working
        callable = callable_dict.get("func_builder",
                                     "key1", "key2",
                                     in_func_key1="key3",
                                     in_func_key2="key3")

    assert callable("local_expired") == dict(
        in_func_arg1="key1",
        in_func_arg2="key2",
        in_func_key1="key3",
        in_func_key2="key3",
        args=("local_expired",),
        kwargs={}, some_global=1,
    )
    # checking redis when local expired
    assert str(caplog.text).count("🍄") == 1

    callable_dict.delete("func_builder",
                         "key1", "key2",
                         in_func_key1="key3",
                         in_func_key2="key3")

    # change the global variable in redis
    # WE DO THIS TEST BECAUSE WE WANT IT TO RELOAD THE FUNCTION
    # WHEN THE GLOBAL VARIABLE CHANGES OR CONFIG IN DATABASE CHANGES

    redis_client.set("some_global", 2, ex=10)
    sleep(1.1)
    for i in range(3):
        callable = callable_dict.get("func_builder",
                                     "key1", "key2",
                                     in_func_key1="key3",
                                     in_func_key2="key3")

    assert callable("test_global") == dict(
        in_func_arg1="key1",
        in_func_arg2="key2",
        in_func_key1="key3",
        in_func_key2="key3",
        args=("test_global",),
        kwargs={},
        some_global=2,
    )
    assert str(caplog.text).count("✨") == 2
    assert str(caplog.text).count("🍄") == 2



from datetime import datetime
from typing import Dict

import aiohttp

from .config import (
    AZURE_SPEECH_KEY, AZURE_TTS_ISSUETOKEN_ENDPOINT)


async def get_token():
    headers = {
        "Ocp-Apim-Subscription-Key": AZURE_SPEECH_KEY,
        "Content-type": "application/x-www-form-urlencoded",
    }
    async with aiohttp.ClientSession() as session:
        async with session.post(
            AZURE_TTS_ISSUETOKEN_ENDPOINT,
                headers=headers) as response:
            return await response.text()


class TokenDeposit:
    """
    Since the API requires to refresh token every 10 minutes.
        This is a handler that can store token for 9 minutes, if time's up
        We'll refersh the token automatically
    """
    def __init__(self):
        self.token = None
        self.expires = 540  # 10 minutes
        self.last_fetch = None

    async def __call__(self) -> str:
        if self.last_fetch is None or \
                (datetime.now() - self.last_fetch).seconds > self.expires:
            self.token = await get_token()
            self.last_fetch = datetime.now()
        return self.token


token_deposit = TokenDeposit()


async def auth_header() -> Dict[str, str]:
    """
    Get auth header for HTTP request
    """
    token = await token_deposit()
    headers = {
        "Ocp-Apim-Subscription-Key": AZURE_SPEECH_KEY,
        "Authorization": f"Bearer {token}",
        }
    return headers



import os
import logging


try:
    AZURE_REGION = os.environ['AZURE_REGION']
    AZURE_TTS_ISSUETOKEN_ENDPOINT = os.environ['AZURE_TTS_ISSUETOKEN_ENDPOINT']
    AZURE_SPEECH_KEY = os.environ['AZURE_SPEECH_KEY']


    AZURE_TTS_VOICE_LIST = f"https://{AZURE_REGION}.tts.speech.microsoft.com/cognitiveservices/voices/list"
    AZURE_TTS_SERVICE = f"https://{AZURE_REGION}.tts.speech.microsoft.com/cognitiveservices/v1"

except KeyError as e:
    logging.error("Please setup env variable 1st")
    raise e


AUDIO_FORMATS = [
    "amr-wb-16000hz",
    "audio-16khz-16bit-32kbps-mono-opus",
    "audio-16khz-32kbitrate-mono-mp3",
    "audio-16khz-64kbitrate-mono-mp3",
    "audio-16khz-128kbitrate-mono-mp3",
    "audio-24khz-16bit-24kbps-mono-opus",
    "audio-24khz-16bit-48kbps-mono-opus",
    "audio-24khz-48kbitrate-mono-mp3",
    "audio-24khz-96kbitrate-mono-mp3",
    "audio-24khz-160kbitrate-mono-mp3",
    "audio-48khz-96kbitrate-mono-mp3",
    "audio-48khz-192kbitrate-mono-mp3",
    "ogg-16khz-16bit-mono-opus",
    "ogg-24khz-16bit-mono-opus",
    "ogg-48khz-16bit-mono-opus",
    "raw-8khz-8bit-mono-alaw",
    "raw-8khz-8bit-mono-mulaw",
    "raw-8khz-16bit-mono-pcm",
    "raw-16khz-16bit-mono-pcm",
    "raw-16khz-16bit-mono-truesilk",
    "raw-22050hz-16bit-mono-pcm",
    "raw-24khz-16bit-mono-pcm",
    "raw-24khz-16bit-mono-truesilk",
    "raw-44100hz-16bit-mono-pcm",
    "raw-48khz-16bit-mono-pcm",
    "webm-16khz-16bit-mono-opus",
    "webm-24khz-16bit-24kbps-mono-opus",
    "webm-24khz-16bit-mono-opus",
]


"""
Manage GPT LLM (Casual language modeling completion tasks)
    to make the query more reliable and trackable
"""


import os
from typing import Any, Dict, List, Optional
import logging
import json


class OpenAIComplete:
    """
    OpenAI Complete wrapper
    A handler to call LLM completion model

    We might have other LLM completion models in the future,
        if there're other providers
    """
    def __init__(
        self,
        api_key: Optional[str] = None,
        model_max_tokens: int = 4095,
    ):
        if api_key is None:
            api_key = os.environ.get('OPENAI_API_KEY')
        if api_key is None:
            raise KeyError(
                "Please set up OPENAI_API_KEY"
            )
        # we import these packages here to avoid
        # import error when we don't have openai
        try:
            import openai
            from transformers import AutoTokenizer
        except ImportError as e:
            logging.error(
                "Please install openai and transformers"
            )
            raise e
        self.openai = openai
        self.openai.api_key = api_key
        self.tokenizer = AutoTokenizer.from_pretrained(
            "gpt2"
        )
        self.model_max_tokens = model_max_tokens

    def cap_tokens(
        self, text: str,
        ask_max_tokens: int = 3090,
    ) -> str:
        """
        Cap the number of tokens in the text
        To avoid the model from generating too long text
        """
        tokens = self.tokenizer.tokenize(text)
        if len(tokens) > ask_max_tokens:
            return self.tokenizer.convert_tokens_to_string(
                tokens[-ask_max_tokens:])
        return text

    def log_io(self, prompt: str, reply: str, **kwargs):
        """
        Log the input and output of the model API
            Feel free to overwrite this function while inheriting
        """
        logging.info(
            json.dumps(
                dict(
                    prompt=prompt,
                    reply=reply,
                    **kwargs
                )
            )
        )

    def __call__(self, prompt: str, **kwargs):
        # if answer's max_tokens is set,
        # we need to cap the prompt
        if "max_tokens" in kwargs:
            prompt = self.cap_tokens(
                prompt,
                self.model_max_tokens - kwargs["max_tokens"])
        reply = self.openai.Completion(
            prompt,
            **kwargs
        ).choices[0].text
        self.log_io(prompt, reply, **kwargs)
        return reply


class Line:
    """
    A line in a chat.
    A line can be of type human or bot.
    line.text is the text of the line.
    line.last_line is the previous line in the chat.
    line.next_line is the next line in the chat.
    """

    def __init__(
        self,
        text: str,
        line_type: str = "unkown",
        last_line: Optional["Line"] = None,
        next_line: Optional["Line"] = None,
    ):
        self.text = text
        if line_type not in ["unkown", "human", "bot"]:
            raise ValueError(
                "line_type must be one of 'unkown', 'human', 'bot'")
        self.line_type = line_type
        self.last_line = last_line
        self.next_line = next_line
        self.data = dict()

    def __repr__(self) -> str:
        return f"Text({self.line_type}): {self.text}"

    def __str__(self) -> str:
        return self.text

    def trace_back_lines(
        self,
        n: Optional[int] = None,
        type_filter: Optional[str] = None,
    ) -> List["Line"]:
        """
        Trace back n lines of type type_filter.
        If n is None, trace back all lines of type type_filter.
        If type_filter is None, trace back all lines.
        """
        if n == 0:
            return []

        if self.last_line is None:
            return []

        # if we turn on the type filter, we need to check if the last text
        # is of the right type
        if type_filter is not None \
                and self.last_line.line_type != type_filter:
            return self.last_line.trace_back_lines(
                n=n, type_filter=type_filter)
        if n is None:
            return [self.last_line] + \
                self.last_line.trace_back_lines(
                    n=None, type_filter=type_filter
            )
        else:
            return [self.last_line] + \
                self.last_line.trace_back_lines(
                    n=n - 1, type_filter=type_filter
            )

    def trace_back_lines_text(
        self,
        n: Optional[int] = None,
        type_filter: Optional[str] = None,
    ) -> List[str]:
        """
        Trace back n lines of type type_filter.
        If n is None, trace back all lines of type type_filter.
        If type_filter is None, trace back all lines.
        """
        return [line.text for line in self.trace_back_lines(
            n=n, type_filter=type_filter
        )]


class ChatBase:
    """
    Base class for chat

    Please inherit this class to implement your own chat
    You have to:
        * set `def __getitem__(self, line: Line) -> List[Dict[str, Any]]`
            to map the line to the references
        * set
            ```python
            def render_reference(
                self,
                ask: Line,
                reference: Dict[str, Any]
            ) -> str:
            ```
            to render the references and the to actual prompt
    """
    def __init__(
        self,
        last_line: Optional[Line] = None,
        complete: Optional[OpenAIComplete] = None,
    ):
        self.last_line = last_line
        if complete is None:
            complete = OpenAIComplete()
        self.complete = complete

    def update_last_line(
        self,
        new_line: Line
    ) -> Line:
        """
        Update the last line of the chat
        Build the connection between new_line and last_line
        """
        if self.last_line is not None:
            # bound the last line to the new line
            new_line.last_line = self.last_line
            self.last_line.next_line = new_line
        self.last_line = new_line
        return self.last_line

    def __getitem__(
        self, line: Line
    ) -> List[Dict[str, Any]]:
        """
        The function we have to map chat with references

        **NOTICE**
        You can leverage `self` to extract the previews data
        self.lines were the preview data
        """
        raise NotImplementedError(
            "You have to define mapping function to find reference"
        )

    def render_reference(
        self,
        ask: Line,
        references: List[Dict[str, Any]],
    ) -> str:
        """
        The function we have to render the references and the to actual prompt,
            Usually by templating
        """
        raise NotImplementedError(
            "You have to define `render_reference` function"
        )

    def __call__(
        self,
        text: str,
        openai_complete_kwargs: Dict[str, Any]
    ) -> str:
        """
        The core function to get the reply based on the last query
            we can still have access to the previous query and replies
        """
        ask = Line(text, line_type='human')
        self.update_last_line(ask)

        references = self[ask]
        ask.data['references'] = references

        prompt = self.render_reference(ask, references)
        complete_kwargs = {
            "model": "davinci-text-003",
            "max_tokens": 512,
        }
        complete_kwargs.update(
            openai_complete_kwargs)

        reply = self.complete(
            prompt=prompt,
            **complete_kwargs,
        )

        reply_line = Line(reply, line_type='bot')

        self.update_last_line(reply_line)

        return reply



import pickle
import redis


class RedisClient:
    def __init__(self, redis_host, redis_port):
        self.redis_client = redis.Redis(host=redis_host, port=redis_port)

    ##########
    # Getters
    ##########
    def get(self, key, default=None):
        """
        Retrieve the byte value. If key does not exist, return default.
        """
        cached_instance = self.redis_client.get(key)
        if cached_instance:
            return pickle.loads(cached_instance)
        else:
            return default

    def mget(self, keys, default=None):
        """
        Get the values of multiple keys at once. If the key does not exists,
        default value is returned instead.
        """
        response_list = self.redis_client.mget(keys)
        return [pickle.loads(res) or default for res in response_list]

    def fetch(self, key, func, ttl, *args, **kwargs):
        """
        Wrapper for Redis GET function, to help get the value
        to the key being passed in, if the key does not exist,
        call the callback function being passed in instead
        """
        res = self.redis_client.get(key)
        if not res:
            res = func(*args, **kwargs)
            if res:
                res = pickle.dumps(res)
                self.redis_client.set(key, res, ttl)
        return pickle.loads(res)

    def exists(self, key):
        """
        Return 1 if the key exists.
        Return 0 if the key does not exist.
        """
        return self.redis_client.exists(key)

    ##########
    # Setters
    ##########
    def set(self, key, value, expiry=None):
        """
        Set key to hold the byte value. If key already holds a value,
        it is overwritten, regardless of its type. Any previous time
        to live associated with the key is discarded on successful
        SET operation.
        """
        serialized_value = pickle.dumps(value)
        self.redis_client.set(key, serialized_value, expiry)

    def mset(self, mapping):
        """
        Set multiple {key: value} pairs at once. Existing keys
        are overwritten, ttls are discarded.
        """
        if mapping:
            for key in mapping:
                mapping[key] = pickle.dumps(mapping[key])
            self.redis_client.mset(mapping)

    ##########
    # List Operations
    ##########
    def list_rpush(self, key, values, expiry=None):
        if not values:
            return 0

        if isinstance(values, list):
            serialized_values = [pickle.dumps(value) for value in values]
            self.redis_client.rpush(key, *serialized_values)
        else:
            serialized_value = pickle.dumps(values)
            self.redis_client.rpush(key, serialized_value)

        # Sets the expiry on the entire list, not the elements
        if expiry:
            self.redis_client.expire(key, expiry)

    def list_lpop(self, key):
        """
        Removes and returns the first element of the list stored at key.
        """
        res = self.redis_client.lpop(key)
        if res:
            return pickle.loads(res)
        else:
            return None

    def llen(self, key):
        """
        Returns the length of the list stored at key. If key does not
        exist, it is interpreted as an empty list and 0 is returned.
        An error is returned when the value stored at key is not a list.
        """
        return self.redis_client.llen(key)

    def list_range(self, key, start=0, stop=-1, default=[]):
        range = self.redis_client.lrange(key, start, stop)
        return [pickle.loads(res) or default for res in range]

    ##########
    # Deletion/Expiry
    ##########
    def delete(self, key):
        self.redis_client.delete(key)

    def flushall(self, asynchronous=False):
        """
        Flush all keys - clear cache
        """
        return self.redis_client.flushall(asynchronous)

    def expire(self, key, expiry):
        """
        Set a timeout on key. After the timeout has expired, the key will
        automatically be deleted. A key with an associated timeout is often
        said to be volatile in Redis terminology.
        It is possible to call EXPIRE using as argument a key that already
        has an existing expire set. In this case the time to live of a key
        is updated to the new value.
        Return value:
        1 if the timeout was set.
        0 if key does not exist.
        """
        return self.redis_client.expire(key, expiry)

    def ttl(self, key):
        """
        Returns the remaining time to live of a key that has a timeout.
        This introspection capability allows a Redis client to check how
        many seconds a given key will continue to be part of the dataset.
        The command returns -2 if the key does not exist.
        The command returns -1 if the key exists but has no associated expire.
        """
        return self.redis_client.ttl(key)



from kubernetes import client, config
from forgebox.unpack import Unpack
import pandas as pd
from .utils import calc_delta
from typing import Dict, List


def load_v1() -> client.CoreV1Api:
    config.load_kube_config()
    v1 = client.CoreV1Api()
    return v1


def list_services(
    namespace: str
) -> pd.DataFrame:
    """
    Return the service list within the namespace
    """
    v1 = load_v1()
    endpoint_res = v1.list_namespaced_endpoints(namespace=namespace)
    services = list()
    for ep in endpoint_res.items:
        name, ports, version, create_ts = Unpack(ep.to_dict())(
            ["metadata", "name"],
            ["subsets", 0, "ports"],
            ["metadata", "resource_version"],
            ["metadata", "creation_timestamp"]
        )
        services.append(
            dict(
                name=name,
                ports=ports,
                version=version,
                create_ts=create_ts
            )
        )
    services_df = pd.DataFrame(services)
    services_df['age'] = services_df.create_ts.apply(calc_delta)
    return services_df


def list_pods(
    namespace: str
) -> pd.DataFrame:
    v1 = load_v1()
    k8_res = v1.list_namespaced_pod(namespace=namespace)
    pods = list()
    for item in k8_res.items:
        data = dict(zip(
            ["name", "version", "app", "created_ts",
             "image", "ready", "started"],
            Unpack(item.to_dict())(
                ["metadata", "name"],
                ["metadata", "resource_version"],
                ["metadata", "labels", "app"],
                ["metadata", "creation_timestamp"],
                ["spec", "containers", 0, "image"],
                ["status", "container_statuses", 0, "ready"],
                ["status", "container_statuses", 0, "started"],
            )))
        pods.append(data)
    pods_df = pd.DataFrame(pods)
    pods_df["image_tag"] = pods_df["image"].apply(lambda x: x.split("/")[1])
    pods_df["age"] = pods_df["created_ts"].apply(calc_delta)
    return pods_df


def services_to_imagetags(namespace: str,) -> Dict[str, List[str]]:
    """
    Given a service name, return the image tag
    """
    v1 = load_v1()
    k8_res = v1.list_namespaced_pod(namespace=namespace)
    services = dict()
    for item in k8_res.items:
        name, image = Unpack(item.to_dict())(
            ["metadata", "name"],
            ["spec", "containers", 0, "image"],
        )
        if name not in services:
            services[name] = list()
        services[name].append(image.split("/")[1])
    return services



from pathlib import Path
import important


def root_dir() -> Path:
    """
    root directory of `important` library
    """
    return Path(important.__file__).parent.absolute()


def conf_dir() -> Path:
    """
    Configuration directory of `important` library
    """
    return root_dir() / "conf"


ROOT_DIR = root_dir()
CONF_DIR = conf_dir()



from redis import Redis
from redis.asyncio import Redis as RedisAsync
from important.location import CONF_DIR
from yaml import safe_load
from gallop.config import BaseConfig
from typing import (
    Dict, Any, Optional, Callable
)
import uuid
import json
import logging
import hashlib
from datetime import timedelta, datetime


class RedisController:
    """
    A redis controller to set and retrieve data conveniently
    with complex redis key pattern, according to the configuration
    at redis.yaml
    """

    def __init__(
        self,
        client: Redis,
        conf: Dict[str, Any],
    ):
        """
        client: redis.Redis, we don't rewrite the redis client
        conf: Dict[str, Any], the configuration of the redis controller
            some must keys:
                pattern: str, the pattern of the redis key
                pattern_kwargs: List[str], the kwargs of the pattern
        """
        self.client = client
        self.config = BaseConfig(**conf)
        for key in ["pattern", "pattern_kwargs"]:
            if key not in self.config:
                raise KeyError(f"{key} not in {self.config}")
        self.ttl_seconds = None

    @classmethod
    def from_task(
        cls,
        client: Redis,
        task: str,
        conf_path: Optional[str] = None,
    ) -> "RedisController":
        f"""
        Instantiate RedisController from task name
        task: key name under 'redis_controller' at {CONF_DIR / "redis.yaml"}
        conf_path: path to the redis configuration file,
            we have default options, but 
        """
        conf_path = CONF_DIR/"redis.yaml" if conf_path is None else conf_path
        with open(conf_path, "r") as f:
            conf = safe_load(f)
        if task in conf['redis_controller']:
            config = conf['redis_controller'][task]
            return cls(client, config)
        else:
            logging.warning(f"🌀 Task {task} not found in redis.yaml")
            raise KeyError(f"Task {task} not found in redis.yaml")

    def __repr__(self,) -> str:
        return f"RedisController({self.config.pattern})"

    @staticmethod
    def ttl_to_seconds(**time_kwargs) -> int:
        """
        Convert ttl key value pairs to seconds
        """
        seconds = 0
        if 'months' in time_kwargs:
            seconds += time_kwargs['months']*30*24*60*60
        if 'days' in time_kwargs:
            seconds += time_kwargs['days']*24*60*60
        if 'hours' in time_kwargs:
            seconds += time_kwargs['hours']*60*60
        if 'minutes' in time_kwargs:
            seconds += time_kwargs['minutes']*60
        if 'seconds' in time_kwargs:
            seconds += time_kwargs['seconds']

        # check if any extra keys
        for key in time_kwargs:
            if key not in ['months', 'days', 'hours', 'minutes', 'seconds']:
                logging.error("supported keys are: months, days, hours, " +
                              "minutes, seconds")
                raise KeyError(f"❓ ttl_to_seconds: unknown key {key}")
        return seconds

    def mk_key(self, **kwargs) -> str:
        """
        Make key from kwargs.
        """
        try:
            return self.config.pattern.format(**kwargs)
        except KeyError as e:
            logging.error(f"❓ 🔑 KeyError: {e}, please input keys as")
            logging.error(self.config["pattern_kwargs"])
            raise e

    def __setitem__(self, kwargs: Dict[str, Any], value: str):
        """
        Set value to redis
        According to the pattern in config
        """
        key = self.mk_key(**kwargs)
        self.client.set(key, value, ex=self.ttl_seconds)
        return key

    def __getitem__(self, kwargs: Dict[str, Any]) -> Optional[str]:
        """
        Get value from redis
        According to the pattern in config
        """
        key = self.mk_key(**kwargs)
        value = self.client.get(key)
        if value is not None:
            return value.decode()
        return None


class RedisControllerAsync(RedisController):
    """
    Async version of RedisController
    """

    def __init__(self,
                 client: RedisAsync,
                 conf: Dict[str, Any],
                 ):
        """
        client: redis.Redis, we don't rewrite the redis client
        conf: Dict[str, Any], the configuration of the redis controller
            some must keys:
                pattern: str, the pattern of the redis key
                pattern_kwargs: List[str], the kwargs of the pattern
        """
        self.client = client
        self.config = BaseConfig(**conf)
        for key in ["pattern", "pattern_kwargs"]:
            if key not in self.config:
                raise KeyError(f"{key} not in {self.config}")
        if 'ttl' in self.config:
            self.ttl_seconds = self.ttl_to_seconds(**self.config.ttl)
        else:
            self.ttl_seconds = None

    async def __aenter__(self):
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        pass

    async def set(self, kwargs: Dict[str, Any], value: str):
        """
        Set value to redis
        According to the pattern in config
        """
        key = self.mk_key(**kwargs)
        await self.client.set(key, value, ex=self.ttl_seconds)
        return key

    async def __getitem__(self, kwargs: Dict[str, Any]) -> Optional[str]:
        """
        Get value from redis
        According to the pattern in config
        """
        key = self.mk_key(**kwargs)
        value = await self.client.get(key)
        if value is not None:
            return value.decode()
        return None


class CallableDictMixin:
    def __repr__(self, ) -> str:
        return f"CallableDictMixin({self.class_map.keys()})"

    def register(self, name: str, cls: Callable):
        """
        Register a class, function
        """
        self.class_map[name] = cls

    def hexify_conf(self, *args, **kwargs):
        """
        Hexify the configuration
        """
        data = dict(
            args=args,
            kwargs=kwargs,
        )
        data_string = json.dumps(data)

        return self.hasher(
            data_string.encode()).hexdigest()[0:8]

    def mk_func_pk(
        self,
        class_name: str,
        *args, **kwargs,
    ):
        """
        Make a function primary key
        """
        function_pk = f"{class_name}:{self.hexify_conf(*args, **kwargs)}"
        return function_pk

    def mk_func_pk_1st(
        self,
        class_name: str,
        *args, **kwargs,
    ):
        """
        Make a function primary key
        """
        function_pk = f"{class_name}:{self.hexify_conf(*args, **kwargs)}"
        if function_pk not in self.function_pks:
            self.function_pks[function_pk] = dict(
                class_name=class_name,
                args=args,
                kwargs=kwargs,
            )
        return function_pk

    def __setitem__(
            self, function_pk: str, callable: Callable):
        """
        Set value to redis
        According to the pattern in config
        """
        self.callables[function_pk] = callable

    def set(self, class_name: str, *args, **kwargs) -> Callable:
        """
        Set the callable
        """
        function_pk = self.mk_func_pk_1st(class_name, *args, **kwargs)
        return self[function_pk]


class CallableDict(RedisController, CallableDictMixin):
    """
    A class to store callable object in redis
    """

    def __init__(
        self,
        client: Redis,
        class_map: Dict[str, Callable] = dict(),
        delay: int = None,
        ttl: Dict[str, int] = dict(
            minutes=2, seconds=30,
        ),
        conf: Optional[Dict[str, Any]] = None,
    ):
        """
        client: redis.Redis, we don't rewrite the redis client
        class_map: Dict[str, Callable], the class map
        delay: int, the span of seconds we don't recheck with redis
        ttl: Dict[str, int], the ttl of redis key
        conf: Optional[Dict[str, Any]],
            the configuration of the redis controller
        """
        if conf is None:
            base_conf = BaseConfig.from_yaml(CONF_DIR/"redis.yaml")
            conf = base_conf["redis_controller"]["callable_dict"]
        super().__init__(client, conf)
        self.class_map = class_map
        self.callables = dict()
        self.function_pks = dict()
        self.uuid_pk_map = dict()
        self.delay_map = dict()
        self.hasher = hashlib.sha256
        ttl = BaseConfig(**ttl)
        self.ttl_seconds = self.ttl_to_seconds(**ttl)
        self.delay = delay

    def reload_callable(self, function_pk: str,) -> Callable:
        function_details = self.function_pks[function_pk]
        class_name = function_details["class_name"]
        args = function_details["args"]
        kwargs = function_details["kwargs"]
        class_ = self.class_map[class_name]
        callable = class_(*args, **kwargs)
        self.callables[function_pk] = callable
        return callable

    def reload(self, function_pk: str,) -> Callable:
        """
        Reload the callable, also register the uuid
        """
        logging.warning(f"✨ Initializing: {function_pk}")
        callable = self.reload_callable(function_pk)
        uuid_pk = str(uuid.uuid4())[:8]

        # set local expire with delay
        if self.delay is not None:
            local_expire = datetime.now() + \
                timedelta(seconds=self.delay)
            self.delay_map[uuid_pk] = local_expire

        self.uuid_pk_map[function_pk] = uuid_pk
        self.client.set(self.mk_key(
            function_pk=function_pk,
            uuid_pk=uuid_pk,
        ), 1, ex=self.ttl_seconds)
        return callable

    def __getitem__(self, function_pk: str) -> Optional[Callable]:
        """
        Get value from redis
        According to the pattern in config
        """
        # uuid is to check destory/ update triggered from redis
        uuid_pk = self.uuid_pk_map.get(function_pk)
        if uuid_pk is None:
            return self.reload(function_pk=function_pk)

        # has uuid but don't have such key in redis
        if self.delay is not None:
            if datetime.now() < self.delay_map.get(uuid_pk, datetime.now()):
                # not checking redis
                return self.callables[function_pk]

        # check redis
        redis_key = self.mk_key(
            function_pk=function_pk,
            uuid_pk=uuid_pk,
        )
        logging.debug("🍄 checking redis")
        if self.client.exists(redis_key) == 0:
            if self.delay is not None:
                if uuid_pk in self.delay_map:
                    del self.delay_map[uuid_pk]
            return self.reload(function_pk=function_pk)

        # prolong delay
        if self.delay is not None:
            self.delay_map[uuid_pk] = datetime.now() + \
                timedelta(seconds=self.delay)
        return self.callables[function_pk]

    def get(self, class_name: str, *args, **kwargs) -> Callable:
        """
        Get the callable
        """
        function_pk = self.mk_func_pk(class_name, *args, **kwargs)
        if function_pk not in self.function_pks:
            # raise KeyError(f"{function_pk} not in {self.function_pks}")
            self.set(class_name, *args, **kwargs)
        return self[function_pk]

    def delete(self, class_name: str, *args, **kwargs):
        """
        Delete the callable
        """
        function_pk = self.mk_func_pk(class_name, *args, **kwargs)
        redis_key = self.mk_key(function_pk=function_pk, uuid_pk="*")
        # delete redis key
        for key in self.client.keys(redis_key):
            logging.warning(f'🗡 [CACHE DELETE]: {key}')
            self.client.delete(key)


class CallableDictAsync(CallableDictMixin):
    def __init__(
        self,
        client: RedisAsync,
        class_map: Dict[str, Callable] = dict(),
        ttl: Dict[str, int] = dict(
            minutes=0,
            seconds=0,
        ),
        delay: int = None,
        conf: Optional[Dict[str, Any]] = None,
    ):
        if conf is None:
            base_conf = BaseConfig.from_yaml(CONF_DIR/"redis.yaml")
            conf = base_conf["redis_controller"]["callable_dict"]
        self.conf = conf
        self.ttl_seconds = RedisController.ttl_to_seconds(**conf["ttl"])
        self.client = client
        self.class_map = class_map
        self.callables = dict()
        self.function_pks = dict()
        self.uuid_pk_map = dict()
        self.delay_map = dict()
        self.hasher = hashlib.sha256
        self.ttl_seconds = RedisController.ttl_to_seconds(**ttl)
        self.delay = delay

    async def reload_callable(self, function_pk: str,) -> Callable:
        function_details = self.function_pks[function_pk]
        class_name = function_details["class_name"]
        args = function_details["args"]
        kwargs = function_details["kwargs"]
        class_ = self.class_map[class_name]
        callable = await class_(*args, **kwargs)
        self.callables[function_pk] = callable
        return callable

    async def reload(self, function_pk: str,) -> Callable:
        """
        Reload the callable, also register the uuid
        """
        logging.warning(f"✨ Initializing: {function_pk}")
        callable = await self.reload_callable(function_pk)
        uuid_pk = str(uuid.uuid4())[:8]

        # set local expire with delay
        if self.delay is not None:
            local_expire = datetime.now() + \
                timedelta(seconds=self.delay)
            self.delay_map[uuid_pk] = local_expire

        self.uuid_pk_map[function_pk] = uuid_pk
        await self.client.set(
            f"callable:{function_pk}:{uuid_pk}",
            1, ex=self.ttl_seconds)
        return callable

    async def get(self, class_name: str, *args, **kwargs) -> Callable:
        """
        Get the callable
        """
        function_pk = self.mk_func_pk_1st(class_name, *args, **kwargs)

        # uuid is to check destory/ update triggered from redis
        uuid_pk = self.uuid_pk_map.get(function_pk)
        if uuid_pk is None:
            return await self.reload(function_pk=function_pk)

        # has uuid but don't have such key in redis
        if self.delay is not None:
            if datetime.now() < self.delay_map.get(uuid_pk, datetime.now()):
                # not checking redis
                return self.callables[function_pk]

        # check redis
        redis_key = f"callable:{function_pk}:{uuid_pk}"

        logging.debug("🍄 checking redis")
        if await self.client.exists(redis_key) == 0:
            if self.delay is not None:
                if uuid_pk in self.delay_map:
                    del self.delay_map[uuid_pk]
            return await self.reload(function_pk=function_pk)

        # prolong delay
        if self.delay is not None:
            self.delay_map[uuid_pk] = datetime.now() + \
                timedelta(seconds=self.delay)
        return self.callables[function_pk]

    async def delete(self, class_name: str, *args, **kwargs):
        """
        Delete the callable
        """
        function_pk = self.mk_func_pk(class_name, *args, **kwargs)
        redis_key = self.mk_key(function_pk=function_pk, uuid_pk="*")

        for key in await self.client.keys(redis_key):
            logging.warning(f'🗡 [CACHE DELETE]: {key}')
            await self.client.delete(key)

    def mk_key(self, **kwargs) -> str:
        """
        Make key from kwargs.
        """
        try:
            return self.conf.pattern.format(**kwargs)
        except KeyError as e:
            logging.error(f"❓ 🔑 KeyError: {e}, please input keys as")
            logging.error(self.conf["pattern_kwargs"])
            raise e



import os
# install the snowflake connector
import snowflake.connector
import pandas as pd
from important.location import CONF_DIR
from typing import Optional
from yaml import safe_load
import logging


def read_sql(sql):
    conn = snowflake.connector.connect(
        user="READ_ONLY_USER",
        password=os.environ['SNOWFLAKE_PASSWORD'],
        account="firework_prod",
        warehouse="COMPUTE_WH",
        database="FIREWORK",
        schema="PUBLIC"
    )
    cur = conn.cursor()
    cur.execute(sql)
    df = cur.fetch_pandas_all()
    conn.close()
    return df


class SnowFlake:
    """
    Snowflake query
    Set SNOFLAKE_PASSWORD in your environment
    Then run SnowFlake.from_config()
    """

    def __init__(self,
                 user: str,
                 account: str,
                 warehouse: str,
                 database: str,
                 schema: str,
                 password: str,
                 ):
        self.user = user
        self.account = account
        self.warehouse = warehouse
        self.database = database
        self.schema = schema
        self.password = password

    def __repr__(self):
        return f"SnowFlake(user={self.user}, account={self.account}, " +\
            f"warehouse={self.warehouse}, schema={self.schema})"

    def __call__(self, sql: str) -> pd.DataFrame:
        conn = snowflake.connector.connect(
            user=self.user,
            password=self.password,
            account=self.account,
            warehouse=self.warehouse,
            database=self.database,
            schema=self.schema)
        cur = conn.cursor()
        cur.execute(sql)
        df = cur.fetch_pandas_all()
        conn.close()
        return df

    @classmethod
    def from_config(
        cls,
        env: str = "prod",
        config_path: Optional[str] = None,
    ) -> "SnowFlake":
        if 'SNOWFLAKE_PASSWORD' not in os.environ:
            raise KeyError(
                "🔑 SNOWFLAKE_PASSWORD not in environment variables")

        if config_path is None:
            config_path = CONF_DIR / "snowflake.yaml"

        with open(config_path, "r") as f:
            conf_data = safe_load(f)

        if env not in conf_data:
            raise KeyError(f"🤔 {env} not in {config_path}")

        conf = conf_data[env]

        return cls(
            password=os.environ['SNOWFLAKE_PASSWORD'],
            **conf)


class SnowFlakeDatabricks(SnowFlake):
    """
    Using with clause in the Snowflake connector
    This will run better on the databricks cluster
    """

    def __call__(self, sql: str) -> pd.DataFrame:
        logging.debug(f"Running SQL: {sql}")
        with snowflake.connector.connect(
                user=self.user,
                password=self.password,
                account=self.account,
                warehouse=self.warehouse,
                database=self.database,
                schema=self.schema) as conn:
            df = pd.read_sql(sql, con=conn)
        logging.debug(
            f"SnowFlake Result:\t{len(df)} lines of data")
        return df



from important.shell.pandas import (
    read_cmd, read_cmd_async
)
import pandas as pd
from subprocess import run
import asyncio
import logging


def list_release(
    namespace: str,
    max_hit: int = 20
) -> pd.DataFrame:
    res = read_cmd(f"helm list -n {namespace} -m {max_hit}")
    if type(res) == str:
        raise ValueError(f"Command execution error: {res}")
    return res


async def list_release_async(
    namespace: str,
    max_hit: int = 20
) -> pd.DataFrame:
    res = await read_cmd_async(f"helm list -n {namespace} -m {max_hit}")
    if type(res) == str:
        raise ValueError(f"Command execution error: {res}")
    return res


def read_release_revisions(
    release: str,
    namespace: str,
    max_hit: int = 20,
) -> pd.DataFrame:
    """
    Read the version list under a release
    """
    command = f"helm history {release} -n {namespace} --max {max_hit}"
    res = read_cmd(command)
    if type(res) == str:
        logging.error(command)
        raise ValueError(f"Command execution error: {res}")
    return res


async def read_release_revisions_async(
    release: str,
    namespace: str,
    max_hit: int = 20,
) -> pd.DataFrame:
    """
    Read the version list under a release
    The async version
    """
    command = f"helm history {release} -n {namespace} --max {max_hit}"
    res = await read_cmd_async(
        command)
    if type(res) == str:
        logging.error(command)
        raise ValueError(f"Command execution error: {res}")
    return res


def rollback_release(
    release: str,
    revision: str,
    namespace: str
) -> pd.DataFrame:
    """
    Rollback a release to a specific revision
    Revision is an integer number, but here we input the string format
    Revision can be found in the output of
    `helm history <release> -n <namespace>`
    """
    command = f"helm rollback {release} {revision} -n {namespace}"
    run_res = run(
        command,
        shell=True,
        capture_output=True,
    )
    if run_res.returncode != 0:
        error_msg = run_res.stderr.decode('utf-8')
        logging.error()
        raise ValueError(f"Command execution error: {error_msg}")

    release_revisions_df = read_release_revisions(release, namespace)
    return release_revisions_df


async def rollback_release_async(
    release: str,
    revision: str,
    namespace: str
) -> pd.DataFrame:
    """
    Rollback a release to a specific revision
    Revision is an integer number, but here we input the string format
    Revision can be found in the output of
    `helm history <release> -n <namespace>`
    The async version of rollback_release
    """
    command = f"helm rollback {release} {revision} -n {namespace}"
    run_res = await asyncio.create_subprocess_shell(
        command,
        stdout=asyncio.subprocess.PIPE,
        stderr=asyncio.subprocess.PIPE,
    )
    stdout, stderr = await run_res.communicate()
    if run_res.returncode != 0:
        error_msg = stderr.decode('utf-8')
        logging.error(command)
        raise ValueError(f"Command execution error: {error_msg}")

    release_revisions_df = await read_release_revisions_async(
        release, namespace)
    return release_revisions_df



from typing import Union
import pandas as pd
from io import StringIO
import logging
from traceback import format_exc
from important.shell.base import shell_async


async def read_cmd_async(
    cmd: str, sep='\t'
) -> Union[pd.DataFrame, str]:
    """
    Parse the command result to a pandas dataframe
    The async version
    cmd: str
        The shell command to run
    sep: str
    Return a dataframe if the command is successful
        and the string can actually be parsed into a dataframe
    Return the string if the command is successful
        but the string can not be parsed into a dataframe
    """
    run_res = await shell_async(cmd)
    if run_res['returncode'] != 0:
        return run_res['stderr']
    value = run_res['stdout']
    try:
        df = pd.read_csv(StringIO(value), sep=sep)
        return df
    except KeyboardInterrupt as e:
        raise e
    except Exception:
        logging.error(format_exc())
        return value



from typing import Dict, Tuple, List
import json


class DAG:
    def __init__(self, pairs: Dict[str, List[str]]) -> None:
        self.pairs = pairs

    @classmethod
    def from_pairs(cls, *args: Tuple[str, str]) -> "DAG":
        pairs = dict()
        for node, next_node in args:
            if node not in pairs:
                pairs[node] = []
            if next_node not in pairs[node]:
                pairs[node].append(next_node)
            if next_node not in pairs:
                pairs[next_node] = []

        return cls(pairs)

    def __len__(self) -> int:
        return len(self.pairs)

    @property
    def nodes(self) -> List[str]:
        return list(self.pairs.keys())

    @property
    def orphans(self) -> List[str]:
        """
        return nodes without parents
        """
        orphans = []
        for node in self.pairs:
            if not any(
                    node in next_nodes
                    for next_nodes in self.pairs.values()):
                orphans.append(node)
        return orphans

    def __getitem__(self, node: str) -> List[str]:
        if node in self.pairs:
            result = [node, ]
            for next_node in self.pairs[node]:
                result += self[next_node]
            return result
        else:
            return []

    def __setitem__(self, node: str, next_node: str) -> None:
        if node not in self.pairs:
            self.pairs[node] = []
        if next_node not in self.pairs[node]:
            self.pairs[node].append(next_node)
        if next_node not in self.pairs:
            self.pairs[next_node] = []

    def to_string(self) -> str:
        return json.dumps(self.pairs, indent=2)

    @classmethod
    def from_string(cls, string: str) -> "DAG":
        pairs = json.loads(string)
        return cls(pairs)



import asyncio
from typing import Dict, Union


async def shell_async(cmd: str) -> Dict[str, Union[int, str]]:
    """
    Run a command asynchronously
    """
    proc = await asyncio.create_subprocess_shell(
        cmd,
        stdout=asyncio.subprocess.PIPE,
        stderr=asyncio.subprocess.PIPE,
    )
    stdout, stderr = await proc.communicate()
    return dict(
        returncode=proc.returncode,
        stdout=stdout.decode('utf-8'),
        stderr=stderr.decode('utf-8'),
    )



# color print for shell

COLORS = {
    "red": "\033[91m",
    "green": "\033[92m",
    "yellow": "\033[93m",
    "blue": "\033[94m",
    "magenta": "\033[95m",
    "cyan": "\033[96m",
    "white": "\033[97m",
    "reset": "\033[0m",
}


def cprint(text: str, color="white", end="\n") -> None:
    """
    Print text in color
    Args:
        text: Text to print
        color: Color to print in, color options are
            red, green, yellow, blue, magenta, cyan, white, reset
        end: End character
    """
    print(COLORS[color] + text + COLORS["reset"], end=end)



import mlflow
from typing import (
    Optional, Union, Dict, Any, List
)
import logging


class MLLogger:
    """
    Make mlflow logging easier
    by setting up the entire handler from experiment id and run_id
    """

    def __init__(
            self, experiment_name: str,
            run_id: Optional[str] = None,
            run_name: Optional[str] = None):
        """
        experiment: name of the experiment
        run_id: if None, create a new run
        """
        self.experiment_name = experiment_name
        self.experiment = mlflow.set_experiment(experiment_name)
        self.experiment_id = self.experiment.experiment_id
        self.run_id = run_id
        if run_id is None:
            # see if we have active run
            if mlflow.active_run() is not None:
                self.run_id = mlflow.active_run().info.run_id
            # set new run
            else:
                run_kwargs = {}
                if run_name is not None:
                    run_kwargs["run_name"] = run_name
                run = mlflow.start_run(
                    experiment_id=self.experiment_id,
                    **run_kwargs
                )
                self.run_id = run.info.run_id
                mlflow.end_run()
        else:
            self.run_id = run_id

    def __add__(self, data: Dict[str, Any]) -> "MLLogger":
        with mlflow.start_run(
            experiment_id=self.experiment_id,
            run_id=self.run_id
        ) as run:
            for k, v in data.items():
                if k in run.data.params:
                    logging.warning(
                        f"Found {k} in params, won't overwrite")
                    continue
                mlflow.log_param(k, v)
        return self

    def __repr__(self) -> str:
        return f"{self.experiment_name}/run:{self.run_id}"

    def __setitem__(self, key: str, value: Any) -> None:
        with mlflow.start_run(
            experiment_id=self.experiment_id,
            run_id=self.run_id
        ) as run:
            if key in run.data.params:
                logging.warning(
                    f"Found {key} in params, won't overwrite")
                return
            mlflow.log_param(key, value)

    def __getitem__(self, key: str) -> Any:
        """
        extract metric or param from run
        """
        with mlflow.start_run(
            experiment_id=self.experiment_id,
            run_id=self.run_id
        ) as run:
            if key in run.data.params:
                return run.data.params[key]
            elif key in run.data.metrics:
                return run.data.metrics[key]
            else:
                raise KeyError(f"{key} not found in params")

    def __mul__(self, data: Dict[str, Union[float, int]]) -> "MLLogger":
        """
        Log metrics
        """
        with mlflow.start_run(
            experiment_id=self.experiment_id,
            run_id=self.run_id
        ) as run:
            for k, v in data.items():
                if k in run.data.metrics:
                    logging.warning(
                        f"Found {k} in metrics, won't overwrite")
                    continue
                mlflow.log_metric(k, v)
        return self

    @property
    def params_data(self) -> Dict[str, Any]:
        with mlflow.start_run(
            experiment_id=self.experiment_id,
            run_id=self.run_id
        ) as run:
            return run.data.params

    @property
    def metrics_data(self) -> Dict[str, Any]:
        with mlflow.start_run(
            experiment_id=self.experiment_id,
            run_id=self.run_id
        ) as run:
            return run.data.metrics

    def query(
        self,
        max_results: int = 10,
        filter_string: Optional[str] = None,
        order_by: List[str] = [],
    ):
        kwargs = dict(order_by=order_by)
        if filter_string is not None:
            kwargs["filter_string"] = filter_string
        return mlflow.search_runs(
            experiment_ids=[self.experiment.experiment_id, ],
            max_results=max_results,
            **kwargs,
        )



import datetime
import os


KAFKA_HOST = os.environ.get("KAFKA_HOST")
TOPIC_NAME = os.environ.get("TOPIC_NAME")
KAFKA_API_KEY = os.environ.get("KAFKA_API_KEY")
KAFKA_API_SECRET = os.environ.get("KAFKA_API_SECRET")


from important.kafka.producer import KafkaProducer


kakfa = KafkaProducer(host=KAFKA_HOST, topic_name=TOPIC_NAME, username=KAFKA_API_KEY, password=KAFKA_API_SECRET)


key = "test_channel_id:123"
value = str(datetime.datetime.utcnow())
kakfa.async_send(key=key, value=value)



from confluent_kafka import Consumer


class KafkaConsumer:
    def __init__(
        self,
        host,
        username,
        password,
        group_id,
        mechanism="PLAIN",
        security_protocol="SASL_SSL",
        auto_offset_reset="earliest",
    ):
        config = self._set_conn_config(
            host, group_id, username, password, mechanism, security_protocol, auto_offset_reset
        )
        self.consumer = Consumer(config)
        self.topics = []

    @staticmethod
    def _set_conn_config(host, group_id, username, password, mechanism, security_protocol, auto_offset_reset):
        """
        Method used to construct connection required attributes.
        This method handles configuration differently if the broker host is
        determined to be a local server.
        """
        config = {"bootstrap.servers": host, "group.id": group_id, "auto.offset.reset": auto_offset_reset}
        if not any(["localhost" in host, "172.0.0.1" in host]):
            config.update(
                {
                    "sasl.username": username,
                    "sasl.password": password,
                    "sasl.mechanism": mechanism,
                    "security.protocol": security_protocol,
                }
            )
        return config

    def subscribe_topic(self, topic_name):
        self.topics.append(topic_name)
        self.consumer.subscribe(self.topics)

    def listen(self, timeout, func, **kwargs):
        try:
            while True:
                msg = self.consumer.poll(timeout=float(timeout))
                if msg:
                    if msg.error():
                        print(msg.error())
                        print("ERROR: %s".format(msg.error()))
                    else:
                        func(msg.topic(), msg.key().decode("utf-8"), msg.value().decode("utf-8"), **kwargs)
        except KeyboardInterrupt:
            self.consumer.close()



from pathlib import Path, PosixPath
from typing import Union, Optional, Dict, Any, List

import aiohttp
import asyncio

from .token import auth_header
from .config import (
    AZURE_SPEECH_KEY, AZURE_TTS_VOICE_LIST, AZURE_TTS_SERVICE)


async def get_voice_list():
    """
    Get voice list

    You can use the following code to analyze the response:
    voices_df = pd.DataFrame(await get_voice_list())
    """
    headers = {"Ocp-Apim-Subscription-Key": AZURE_SPEECH_KEY}
    async with aiohttp.ClientSession() as session:
        async with session.get(
            AZURE_TTS_VOICE_LIST,
                headers=headers) as response:
            print(response.status)
            return await response.json()


class VoiceBank:
    """
    Voice bank that holds the config to all voices
    """
    def __init__(self, locale: Optional[str] = None,
                 voice_list=List[Dict[str, Any]]):
        """
        locale: str, optional
            Locale of the voice, e.g. "en-US", "zh-CN"
        voice_list: list of dict
            List of voice config
            """
        self.language = locale
        self.voice_list = voice_list
        if locale is not None:
            if locale[-1] == "*":
                locale = locale[:-1]
                self.voice_list = [
                    voice for voice in self.voice_list
                    if voice["Locale"].startswith(locale)]
            else:
                self.voice_list = [
                    voice for voice in self.voice_list
                    if voice["Locale"] == locale]

    @classmethod
    async def from_azure(cls, locale: Optional[str] = None):
        """
        Get voice bank from Azure
        """
        voice_list = await get_voice_list()
        return cls(locale, voice_list)

    def __call__(self, voice_name: str) -> Dict[str, Any]:
        """
        Get voice by name
        """
        for voice in self.voice_list:
            if voice["DisplayName"] == voice_name:
                return voice
            elif voice["ShortName"] == voice_name:
                return voice
        raise ValueError(f"Voice {voice_name} not found")

    def name_list(self) -> List[str]:
        """
        Get all voice names
        """
        return [voice["DisplayName"] for voice in self.voice_list]

    def __repr__(self):
        allnames = ",".join(
            self.name_list())
        return (
            f"VoiceBank({self.language})" +
            f"[{len(self.voice_list)} voices]:\n" +
            allnames
        )


async def text_to_speech(
        text,
        voice_kwargs,
        style=None,
        output_format="audio-16khz-128kbitrate-mono-mp3",
        user_agent="important"
        ) -> bytes:
    """
    Convert text to speech
    """
    headers = await auth_header()
    headers["Content-Type"] = "application/ssml+xml"
    headers["X-Microsoft-OutputFormat"] = output_format
    headers["User-Agent"] = user_agent

    locale = voice_kwargs.get("Locale", "en-US")
    gender = voice_kwargs.get('Gender')
    short_name = voice_kwargs.get('ShortName')

    style = "" if style is None else f"style='{style}'"

    ssml = f"""<speak version='1.0' xml:lang='{locale}'>
    <voice xml:lang='{locale}' xml:gender='{gender}' {style}
        name='{short_name}'>{text}</voice></speak>"""
    async with aiohttp.ClientSession() as session:
        async with session.post(
            AZURE_TTS_SERVICE, headers=headers, data=ssml
                ) as response:
            if response.status // 100 == 2:
                return await response.read()
            else:
                raise ConnectionError(
                    f"🔌 CODE:{response.status}, {await response.text()}")


def save_audio_bytes(
        data: bytes,
        directory: Union[PosixPath, str],
        file_stem: str,
        output_format="audio-16khz-128kbitrate-mono-mp3",) -> str:
    """
    Save audio bytes to file
    """
    filename = f"{file_stem}.{output_format.split('-')[-1]}"
    directory = Path(directory)
    directory.mkdir(
        parents=True, exist_ok=True
    )
    save_path = directory / filename
    with open(save_path, 'wb') as f:
        f.write(data)
    return save_path



from confluent_kafka import Producer


class KafkaProducer:
    def __init__(
        self,
        host,
        topic_name,
        username,
        password,
        mechanism="PLAIN",
        security_protocol="SASL_SSL",
        flushing_interval_secs=1.0,
        verbose=1.0,
    ):
        self.topic_name = topic_name
        self.flushing_interval_secs = flushing_interval_secs
        self.verbose = verbose

        config = self._set_conn_config(host, username, password, mechanism, security_protocol)

        self.producer = Producer(config)

    @staticmethod
    def _set_conn_config(host, username, password, mechanism, security_protocol):
        """
        Method used to construct connection required attributes.
        This method handles configuration differently if the broker host is
        determined to be a local server.
        """
        config = {"bootstrap.servers": host}
        print(host)
        if not any(["localhost" in host, "172.0.0.1" in host]):
            config.update(
                {
                    "sasl.username": username,
                    "sasl.password": password,
                    "sasl.mechanism": mechanism,
                    "security.protocol": security_protocol,
                }
            )
        return config

    @staticmethod
    def _delivery_callback(err, msg):
        if err:
            print("ERROR: Message failed delivery: {}".format(err))
        else:
            print(
                "Produced event to topic {topic}: key = {key:12} value = {value:12}".format(
                    topic=msg.topic(), key=msg.key().decode("utf-8"), value=msg.value().decode("utf-8")
                )
            )

    def _add_msg_to_producer_buffer(self, key, value):
        cb = self._delivery_callback if self.verbose > 0 else None
        self.producer.produce(topic=self.topic_name, key=key, value=value, callback=cb)

    def send(self, key, value):
        self._add_msg_to_producer_buffer(key, value)
        self.producer.flush()

    def async_send(self, key, value):
        self._add_msg_to_producer_buffer(key, value)
        self.producer.poll(self.flushing_interval_secs)



from datetime import datetime, timedelta


def calc_delta(time: datetime) -> timedelta:
    """
    Calculate how much time has passed since the time
    """
    return datetime.now()-(
        datetime
        .fromtimestamp(time.timestamp())
    )



import os

from important.kafka.consumer import KafkaConsumer

KAFKA_HOST = os.environ.get("KAFKA_HOST")
KAFKA_API_KEY = os.environ.get("KAFKA_API_KEY")
KAFKA_API_SECRET = os.environ.get("KAFKA_API_SECRET")
GROUP_ID = os.environ.get("GROUP_ID")
TOPIC_NAME = os.environ.get("TOPIC_NAME")
POLLING_TIMEOUT_SECS = os.environ.get("POLLING_TIMEOUT_SECS")


# We can use this as the control plane for the various kafka-based functionalities
# so the actual listener method stays simple and agnostic to the usecase
def kafka_consumer_controller(topic, msg_key, msg_value, **kwargs):
    if msg_key == "test_channel_id:123":
        print("123")
    else:
        print("abc")


# Create Consumer instance
consumer = KafkaConsumer(KAFKA_HOST, KAFKA_API_KEY, KAFKA_API_SECRET, GROUP_ID)
consumer.subscribe_topic(TOPIC_NAME)
consumer.listen(POLLING_TIMEOUT_SECS, kafka_consumer_controller)



from subprocess import run
from datetime import datetime
import uuid
import json
from pathlib import Path


def add_to_file(s3path, data):
    filename = (
        f"{datetime.now().strftime('%Y%m%d%H%M%S')}"
        f"-{uuid.uuid4()[:8]}.json"
        )
    temp_file = f"/tmp/{filename}"
    with open(temp_file, 'w') as f:
        f.write(json.dumps(data, indent=4,))

    run_res = run(
        f"aws s3 cp {temp_file} {Path(s3path) / filename}",
        shell=True)

    if run_res.returncode != 0:
        raise ValueError("Logging pushing error")



import asyncio
from typing import Dict, Any
import json


async def ecr_list_images(
    repo_name: str
) -> Dict[str, Any]:
    """
    List all images in ECR
    """
    command = (
        "aws ecr list-images "
        f"--repository-name {repo_name}"
    )

    proc = await asyncio.subprocess.create_subprocess_shell(
        command,
        stdout=asyncio.subprocess.PIPE,
        stderr=asyncio.subprocess.PIPE,
    )

    stdout, stderr = await proc.communicate()

    stdout = stdout.decode('utf-8')
    stderr = stderr.decode('utf-8')

    return_code = proc.returncode

    if return_code != 0:
        raise ValueError(f"Command execution error: {stderr}")

    try:
        data = json.loads(stdout)
        return data
    except json.JSONDecodeError:
        data = dict(
            returncode=proc.returncode,
            stdout=stdout,
            stderr=stderr,
        )
        return data



import re
from typing import Any, Dict, List, Optional, Union
import logging

from pydantic import BaseModel
import tiktoken
import openai

from .config import (
    MODEL_TOKEN_CONFIGS, OPENAI_API_KEY
    )


class TokenLengthGuard:
    """
    A class to guard against token length,
    to prevent from hitting the openai api limit
    """
    def __init__(
        self,
        model: str,
        hard_roof: int,
        preserve_preference: str = "head",
        safe_buffer: int = 16,
        force_assistant_muffle: bool = False,
    ):
        """
        model: name of the model we're going to connect
        hard_roof:
            the maximum length of tokens
            combining the question and answer
        preserve_preference:
            head or tail, which part of the tokens to preserve
        safe_buffer:
            the buffer to prevent from hitting the hard roof
            in case our tokenization is not entirely accurate
        force_assistant_muffle:
            whether to force the muffle the content from role: assistant
                (take out the middle part of the paragraph)
        """
        self.hard_roof = hard_roof
        assert preserve_preference in ["head", "tail"], \
            "preserve_preference must be one of head, tail"
        self.preserve_preference = preserve_preference
        self.safe_buffer = safe_buffer
        self.model = model
        # load the official tokenize to assimulate the openai tokenizing count
        self.tokenizer = tiktoken.encoding_for_model(model)
        self.force_assistant_muffle = force_assistant_muffle

    def slicing(
        self, tokens: List[str], remain_length: int = 512
    ) -> List[str]:
        """
        Slice the tokens to the desired length
        Could be cutting from head or tail
        """
        if self.preserve_preference == "head":
            return tokens[:remain_length]
        else:
            return tokens[-remain_length:]

    def guard_prompt(
        self, prompt: int,
        min_answer_tokens: int = 512
    ) -> Dict[str, Union[int, str]]:
        """
        Guard the prompt length
        """
        token_ids = self.tokenizer.encode(prompt)
        original_length = len(token_ids)
        if original_length > self.hard_roof - min_answer_tokens:
            logging.debug(
                f"[CUT]{original_length} cut to "
                f"preserve `{self.preserve_preference}`"
            )
            sliced = self.slicing(
                token_ids,
                remain_length=self.hard_roof - min_answer_tokens-self.safe_buffer)
            prompt = self.tokenizer.decode(
                sliced
            )
            return {
                "prompt": prompt,
                "max_tokens":
                    self.hard_roof - len(sliced) - self.safe_buffer,
            }
        else:
            return {
                "prompt": prompt,
                "max_tokens":
                    self.hard_roof - original_length - self.safe_buffer,
            }

    def sum_msg_len(self, messages: List[Dict[str, str]]) -> int:
        """
        Calculate the total length of message length
        According to Ray's reverse engineering result on
        how the chatgpt API calculate the same number
        ```latex
        N = \Sigma(Content_{i} + 5) + 2
        ```
        """
        return sum(
            map(
                lambda x: len(x) + 5,
                self.tokenizer.encode_batch(
                    list(i['content'] for i in messages)
                )
            )
        ) + 2

    @staticmethod
    def select_role(
        messages: List[Dict[str, str]],
        role: str
    ) -> List[Dict[str, str]]:
        """
        Filter the messages by role
        """
        if role not in ['user', 'assistant', 'system']:
            raise ValueError(f"Role `{role}` not supported")
        return list(filter(lambda x: x['role'] == role, messages))

    def body_muffle(
        self, text: str,
    ):
        """
        Muffle long text's body, leave the head and tail sentences
        """
        # split sentence by '\n' or '.' or '?'
        sentences = re.split(r'[\n\.?]', text.strip("\n").strip("."))
        if len(sentences) < 3:
            return text
        else:
            return "\n".join(
                [sentences[0], sentences[-1]]
            )

    def iter_body_muffle_a_role(
        self, messages: Dict[str, str], role: str
    ) -> Dict[str, str]:
        """
        Muffle the body of a role
        """
        if role == 'user':
            iter_list = messages[:-1]
        else:
            iter_list = messages
        for msg in iter_list:
            if msg['role'] == role:
                original_length = len(
                    self.tokenizer.encode(msg['content'])) + 5
                msg['content'] = self.body_muffle(msg['content'])
                token_length_diff = original_length - (len(
                    self.tokenizer.encode(msg['content'])) + 5)
                yield msg, token_length_diff
            else:
                yield msg, 0

    def prepare_result(
        self,
        messages: List[Dict[str, str]],
        total: int,
        strategy: Optional[str] = None
    ) -> List[Dict[str, Any]]:
        """
        Prepare the return result
        """
        result = dict(
            messages=messages,
            max_tokens=self.hard_roof - total - self.safe_buffer,
            prompt_length=total,
        )
        if strategy is not None:
            result.update(dict(strategy=strategy))
        return result

    def guard_messages(
        self, messages: List[Dict[str, str]],
        min_answer_room: int = 512,
    ) -> List[Dict[str, str]]:
        """
        Guard the messages length
            this input format is for chat completion
        """
        # check the validation for min_answer_room
        if min_answer_room + self.safe_buffer > self.hard_roof:
            raise ValueError(
                "min_answer_room + safe_buffer > hard_roof, "
                "please adjust the parameters"
            )

        if self.force_assistant_muffle:
            messages = list(msg for msg, _
                             in self.iter_body_muffle_a_role(
                                messages, 'assistant')
                                )
        total = self.sum_msg_len(messages)
        if total <= self.hard_roof - min_answer_room - self.safe_buffer:
            return self.prepare_result(
                messages, total, strategy=(
                    "force_assitant_muffle"
                    if self.force_assistant_muffle else "no muffle"))

        # still too long
        # try muffle body for assistant role
        if self.force_assistant_muffle is False:
            for msg, diff in self.iter_body_muffle_a_role(
                    messages, 'assistant'):
                total -= diff
                if total <= (
                        self.hard_roof
                        - min_answer_room - self.safe_buffer):
                    return self.prepare_result(
                        messages, total,
                        strategy="1by1_assistant_muffle")

        # still too long
        # try to muffle body for user role
        for msg, diff in self.iter_body_muffle_a_role(
                messages, 'user'):
            total -= diff
            if total <= (
                    self.hard_roof
                    - min_answer_room - self.safe_buffer):
                return self.prepare_result(
                    messages, total,
                    strategy="1by1_user_muffle")

        total = self.sum_msg_len(messages)

        # still too long trucate assistant
        messages = list(
            msg for msg
            in messages if msg['role'] != 'assistant')

        total = self.sum_msg_len(messages)
        if total <= self.hard_roof - min_answer_room - self.safe_buffer:
            return self.prepare_result(
                messages, total,
                "no_assistant")

        # still too long
        # save last user input, remove all other user input
        last_input = messages[-1]
        messages = list(
            msg for msg in messages[:-1]
            if msg['role'] != "user") + [last_input, ]

        total = self.sum_msg_len(messages)
        if total <= self.hard_roof - min_answer_room - self.safe_buffer:
            return self.prepare_result(
                messages, total,
                "only_last_user_input")

        # still too long
        # muffle the body of the last user input
        last_msg = messages[-1]
        last_msg['content'] = self.body_muffle(last_msg['content'])
        total = self.sum_msg_len(messages)
        if total <= self.hard_roof - min_answer_room - self.safe_buffer:
            return self.prepare_result(
                messages, total,
                "muffle_last_user_input"
                )

        # still too long

        raise ValueError(
            "The message is still too long to be processed, "
            "we've tried all the strategies: "
            "we even removed any answer from the model"
            "we kept only last user input, "
            "we kept only the 1st and the last sentence "
            "of the last user input. "
            f"current total token length: {total}, "
            f"hard roof for all the tokens: {self.hard_roof}, "
            f"minimun room for answer tokens: {min_answer_room}, "
            f"safe buffer tokens: {self.safe_buffer}"
        )


class ChatLine(BaseModel):
    content: str
    role: str = "user"


class ChatConnection:
    kwargs_filter = [
        "model",
        "temperature",
        "top_p",
        "n",
        "stream",
        "stop",
        "presence_penalty",
        "frequency_penalty",
        "logit_bias",
        "user",
        "min_answer_room",
    ]

    def __init__(
        self,
        model: Optional[str] = None,
        openai_api_key: Optional[str] = None,
        system_role: Optional[Union[str, List[str]]] = None,
        min_answer_room: Optional[int] = None,
    ) -> None:
        if model is None:
            model = "gpt-3.5-turbo"
        self.model = model
        if openai_api_key is None:
            if OPENAI_API_KEY == "not set":
                raise KeyError(
                    "Please set OPENAI_API_KEY system env"
                    " or pass in openai_api_key")
            openai_api_key = OPENAI_API_KEY
        openai.api_key = openai_api_key
        if system_role is None:
            system_role = (
                "You are the universal customer service agent "
                "who can answer any knowledge about any product "
                "in detail"
            )
        self.system_role = system_role

        model_config = MODEL_TOKEN_CONFIGS[model]

        if min_answer_room is None:
            min_answer_room = model_config["min_answer_room"]
        self.min_answer_room = min_answer_room

        self.token_guard = TokenLengthGuard(
            model=model,
            hard_roof=model_config["hard_roof"],
            safe_buffer=model_config["safe_buffer"],
        )

    def unpack_system_role(
            self, system_role: Optional[Union[str, List[str]]] = None,
            ) -> List[Dict[str, str]]:
        """
        Make sure system role is set to default format
        """
        if system_role is None:
            system_role = self.system_role
        if isinstance(system_role, list):
            if len(system_role) == 0:
                return []
            elif type(system_role[0]) is str:
                return list(
                    {"role": "system", "content": i}
                    for i in system_role)
            else:
                return system_role
        elif isinstance(system_role, str):
            return [{"role": "system", "content": system_role}]
        else:
            raise ValueError("system_role must be a string or a list")

    def prepare_messages(
            self, texts: List[Dict[str, str]],
            system_role: Optional[Union[str, List[str]]] = None,
            ) -> List[Dict[str, str]]:
        """
        Make sure messages is set to default format
        """
        system_role: List[ChatLine] = self.unpack_system_role(system_role)

        if isinstance(texts, str):
            messages = [dict(role="user", content=texts)]
        elif isinstance(texts, list):
            messages = []
            for msg in texts:
                if isinstance(msg, str):
                    messages.append(dict(role="user", content=msg))
                elif isinstance(msg, dict):
                    if "role" not in msg:
                        msg["role"] = "user"
                    if set(msg.keys()) != {"role", "content"}:
                        raise KeyError(
                            "dict in texts must only have role and content")
                    messages.append(msg)
                else:
                    raise ValueError(
                        "texts must be a list of strings or dicts")
        else:
            raise ValueError("texts must be a string or a list")

        # insert system role back into messages
        if len(system_role) > 0:
            messages = messages[:-1] + system_role + messages[-1:]

        guarded = self.token_guard.guard_messages(
            messages, min_answer_room=self.min_answer_room)
        return guarded

    def __call__(
        self,
        texts: Union[
            str, List[ChatLine]],
        system_role: Optional[Union[str, List[str]]] = None,
        **kwargs,
    ) -> str:
        """
        Send chat request to openapi
        """

        guarded = self.prepare_messages(texts, system_role)

        messages = guarded["messages"]
        max_tokens = guarded["max_tokens"]

        input_kwargs = {
            "model": self.model,
            "messages": messages,
            "max_tokens": max_tokens,
        }

        for k, v in kwargs.items():
            if k in self.kwargs_filter:
                input_kwargs[k] = v

        logging.debug(f"input_kwargs: {input_kwargs}")

        res = openai.ChatCompletion.create(**input_kwargs)

        if "choices" not in res:
            logging.error(f"openai api returned invalid response: {res}")
            raise ConnectionError("openai api returned invalid response")

        return res["choices"][0]["message"]["content"]

    async def acreate(
        self,
        texts: Union[
            str, List[ChatLine]],
        system_role: Optional[Union[str, List[str]]] = None,
        **kwargs,
    ) -> str:
        """
        Send chat request to openapi
        """

        guarded = self.prepare_messages(texts, system_role)

        messages = guarded["messages"]
        max_tokens = guarded["max_tokens"]

        input_kwargs = {
            "model": self.model,
            "messages": messages,
            "max_tokens": max_tokens,
        }

        for k, v in kwargs.items():
            if k in self.kwargs_filter:
                input_kwargs[k] = v

        logging.debug(f"input_kwargs: {input_kwargs}")

        res = await openai.ChatCompletion.acreate(**input_kwargs)

        if "choices" not in res:
            logging.error(f"openai api returned invalid response: {res}")
            raise ConnectionError("openai api returned invalid response")

        return res["choices"][0]["message"]["content"]




import unittest

from important.kafka.producer import KafkaProducer


class TestProducer(unittest.TestCase):

    def setUp(self):
        self.test_topic_name = 'test_topic'
        self.test_username = 'test_username'
        self.test_password = 'test_password'
        self.test_mechnism = 'PLAIN'
        self.test_security_protocol='SASL_SSL'

    def test_connection_config_local(self):
        # Given
        localhost = 'localhost:9092'

        # When
        actual_config = KafkaProducer(host=localhost,
                                      topic_name=self.test_topic_name,
                                      username=self.test_username,
                                      password=self.test_password,)\
            ._set_conn_config(host=localhost,
                              username=self.test_username,
                              password=self.test_password,
                              mechanism=self.test_mechnism,
                              security_protocol=self.test_security_protocol)

        expected_config = {'bootstrap.servers': localhost}

        # Then
        self.assertDictEqual(actual_config, expected_config)

    def test_connection_config_remote(self):
        # Given
        remotehost = 'abc.us-west-2.aws.confluent.cloud:9092'

        # When
        actual_config = KafkaProducer(host=remotehost,
                                      topic_name=self.test_topic_name,
                                      username=self.test_username,
                                      password=self.test_password,)\
            ._set_conn_config(host=remotehost,
                              username=self.test_username,
                              password=self.test_password,
                              mechanism=self.test_mechnism,
                              security_protocol=self.test_security_protocol)

        expected_config = {'bootstrap.servers': remotehost,
                           'sasl.username': self.test_username,
                           'sasl.password': self.test_password,
                           'sasl.mechanism': self.test_mechnism,
                           'security.protocol': self.test_security_protocol}

        # Then
        self.assertDictEqual(actual_config, expected_config)


if __name__ == '__main__':
    unittest.main()



try:
    import kubernetes
    print(
        "kubernetes python SDK: "
        f"version {kubernetes.__version__}"
        )
    from .crud import (
        list_services,
        list_pods,
        services_to_imagetags
    )
except ImportError:
    print("kubernetes python SDK not installed")
    print("Please run `pip install kubernetes`")


try:
    import kubernetes_asyncio
    print(
        "kubernetes_asyncio python SDK: "
        f"version {kubernetes_asyncio.__version__}"
        )
    from .crud_async import (
        list_services_async, list_pods_async,
        services_to_imagetags_async
        )
except ImportError:
    print("kubernetes_asyncio python SDK not installed")
    print("Please run `pip install kubernetes_asyncio`")



import os

from important.redis_cache.redis_client import RedisClient

redis_host = os.getenv("REDIS_HOST")
redis_port = os.getenv("REDIS_PORT", 6379)
redis_client = RedisClient(redis_host, redis_port)

# Test String Set/Get
key = "test_string"
value_in = "hello world"
redis_client.set(key, value_in)
value_out = redis_client.get(key)
assert value_in == value_out
redis_client.delete(key)

# Test Float Set/Get
key = "test_float"
value_in = 5.3
redis_client.set(key, value_in)
value_out = redis_client.get(key)
assert value_in == value_out
redis_client.delete(key)

# Test List Implementation
key = "test_list_key"
redis_client.list_rpush(key, 1)
redis_client.list_rpush(key, 2)
redis_client.list_rpush(key, "3")
assert redis_client.llen(key) == 3
assert redis_client.list_lpop(key) == 1
assert redis_client.list_range(key) == [2, "3"]
redis_client.delete(key)


# test the default fallback in getter
if redis_client.exists("test"):
    redis_client.delete("test")
assert not redis_client.exists("test")
default = "empty"
value_out = redis_client.get("test", default=default)
assert value_out == default


# test the fetch method
def toUpper(s):
    return s.upper()


redis_client.set("key", "value")
assert redis_client.fetch("key", toUpper, None, "lower") == "value"
redis_client.delete("key")

assert redis_client.fetch("empty_key", toUpper, None, "lower") == "LOWER"
redis_client.delete("empty_key")

assert redis_client.fetch("empty_key_num", len, None, "hello") == 5
redis_client.delete("empty_key_num")

# Test mset and mget
redis_client.mset(
    {
        "key1": "1",
        "key2": "2",
        "key3": "3",
    }
)
assert redis_client.mget(["key1", "key2", "key3"]) == ["1", "2", "3"]
redis_client.set("key3", 3)
assert redis_client.mget(["key1", "key2", "key3"]) == ["1", "2", 3]
for k in ["key1", "key2", "key3"]:
    redis_client.delete(k)

print("All checks passed")



import mlflow
from important.logger.mllogger import MLLogger
import logging
from time import sleep


mlflow.set_tracking_uri("http://localhost:9876")


mllogger = MLLogger("test_experiment", run_name="test_run")


def test_logging_params():
    global mllogger
    logging.info("input params")
    mllogger = mllogger + dict(
        a=1,
        b=2,
        c="model-name",
    )

    logging.info("input params with duplicate keys")
    mllogger = mllogger + dict(
        a=4,
        d=3,
    )

    assert mllogger["a"] == "1"
    assert mllogger["b"] == "2"
    assert mllogger["c"] == "model-name"
    assert mllogger["d"] == "3"


def test_logging_metrics():
    global mllogger
    logging.info("input metrics")
    mllogger = mllogger * dict(
        train_loss=0.1,
        val_loss=0.2,
        test_loss=0.3,
    )

    logging.info("input metrics with duplicate keys")
    mllogger = mllogger * dict(
        train_loss=0.4,
        val_loss=0.5,
    )

    assert mllogger["train_loss"] == 0.1
    assert mllogger["val_loss"] == 0.2
    assert mllogger["test_loss"] == 0.3


mllogger2 = MLLogger("test_experiment", run_id=mllogger.run_id)


def test_query():
    df = mllogger2.query()
    assert df["metrics.train_loss"].values[0] == 0.1
